'''
Provides electrical components for pyliso.

All components have a compInstn method which returns the string corresponding
to the LISO command for creating that component
'''
import numpy as np


SIprefix = {
    'f': 1.0e-15,  # femto
    'p': 1.0e-12,  # pico
    'n': 1.0e-9,   # nano
    'u': 1.0e-6,   # micro
    'm': 1.0e-3,   # milli
    'k': 1.0e3,    # kilo
    'M': 1.0e6,    # mega
    'G': 1.0e9,    # giga
     }


def convert_SI_string_to_value(numberString):
    '''Takes in a single string representing a value using letter,
    i.e. "49k" for 49000, and outputs the value as a float '''
    number = numberString
    for prefix in SIprefix.keys():
        if prefix in numberString:
            multiplier = SIprefix[prefix]  # get the value
            number = float(numberString[:-1]) * multiplier
            break
    if type(number) == str:  # no SI prefix
        number = float(numberString)
    return number


def convert_value_to_SI_string(number):
    '''Takes in number, outputs string compatible with LISO'''
    if np.abs(number) < 1.0e-18:
        # if the number is exactly zero, then it is probably
        # a connector between two nodes placed on purpose
        numberString = str(number)

    elif number < 1.0e-15:
        print 'Number', number, ' is too small for LISO SI units'
        numberString = str(number)

    elif number > 1.0e12:
        print 'Number', number, 'is too big for LISO SI units'
        numberString = str(number)

    elif number > 1.0 and number < 1000.0:
        numberString = str(number)
        return numberString

    for prefix in SIprefix.keys():
        prefixValue = SIprefix[prefix]
        testValue = number / prefixValue
        if testValue >= 1.0 and testValue < 1000.0:
            numberString = str(testValue) + prefix
            break
    return numberString


class Resistor:
    def __init__(self, name, value, node1, node2):
        self.name = name
        self.value = value
        self.node1 = node1
        self.node2 = node2

    def __repr__(self):
        return '{:s} is a {:e} Ohm resistor'.format(self.name, self.value)

    def setValue(self, value):
        self.value = value

    def getValue(self):
        return self.value

    def compInstn(self):
        cmd = 'r {:s} {:e} {:s} {:s}\n'.format(
            self.name, self.value, self.node1, self.node2)
        return cmd


class Capacitor:
    def __init__(self, name, value, node1, node2):
        self.name = name
        self.value = value
        self.node1 = node1
        self.node2 = node2

    def __repr__(self):
        return '{:s} is a {:e} F capacitor'.format(self.name, self.value)

    def setValue(self, value):
        self.value = value

    def getValue(self):
        return self.value

    def compInstn(self):
        cmd = 'c {:s} {:e} {:s} {:s}\n'.format(
            self.name, self.value, self.node1, self.node2)
        return cmd


class Inductor:
    def __init__(self, name, value, node1, node2):
        self.name = name
        self.value = value
        self.node1 = node1
        self.node2 = node2

    def __repr__(self):
        return '{:s} is a {:e} H inductor'.format(self.name, self.value)

    def setValue(self, value):
        self.value = value

    def getValue(self):
        return self.value

    def compInstn(self):
        cmd = 'l {:s} {:e} {:s} {:s}\n'.format(
            self.name, self.value, self.node1, self.node2)
        return cmd


class OpAmp:
    def __init__(self, name, typ, nodep, nodem, nodeout, **kwargs):
        self.name = name
        self.typ = typ
        self.nodep = nodep
        self.nodem = nodem
        self.nodeout = nodeout
        self.characterisics = kwargs

    def __repr__(self):
        return '{:s} is a {:s} op-amp'.format(self.name, self.typ)

    def setType(self, value):
        self.typ = value

    def setCharacteristic(self, char, value):
        '''
        Needs work
        '''
        self.characterisics[char] = value

    def getCharacteristic(self, char):
        '''
        Needs work
        '''
        return self.characterisics[char]

    def compInstn(self):
        '''
        Needs work
        '''
        cmd = 'op {:s} {:s} {:s} {:s} {:s}\n'.format(
            self.name, self.typ, self.nodep, self.nodem, self.nodeout)
        for char, val in self.characterisics.items():
            cmd += '{:s}={:e}'.format(char, val)
        return cmd


class MutualInductance:
    def __init__(self, name, value, inductor1, inductor2):
        self.name = name
        self.value = value
        self.inductor1 = inductor1
        self.inductor2 = inductor2

    def __repr__(self):
        return '{:s} is a mutual inductance with coupling {:f}'.format(
            self.name, self.value)

    def setValue(self, value):
        self.value = value

    def getValue(self):
        return self.value

    def compInstn(self):
        cmd = 'm {:s} {:f} {:s} {:s}\n'.format(
            self.name, self.value, self.inductor1, self.inductor2)
        return cmd


class TransmissionLine:
    def __init__(self, name, length, Z, C, R, G, fskin, n3, n4, n5, n6):
        self.name = name
        self.length = length
        self.Z = Z
        self.C = C
        self.R = R
        self.G = G
        self.fskin = fskin
        self.n3 = n3
        self.n4 = n4
        self.n5 = n5
        self.n6 = n6

    def __repr__(self):
        return '{:s} is a transmission line'.format(self.name)

    def compInstn(self):
        cmd = ('tr {:s} {:f} {:f} {:f} {:f} {:f} {:f} '
               '{:s} {:s} {:s} {:s}'.format(
                   self.name, self.length, self.Z, self.C, self.R, self.G,
                   self.fskin, self.n3, self.n4, self.n5, self.n6))
        return cmd


class Pole:
    def __init__(self, freq, Q=None):
        self.freq = freq
        self.Q = Q

    def __repr__(self):
        if self.Q:
            return 'a c.c. pole pair at {:f} Hz with Q={:f}'.format(
                self.freq, self.Q)
        else:
            return 'a single pole at {:f} Hz'.format(self.freq)

    def compInstn(self):
        if self.Q:
            return 'pole {:f} {:e}\n'.format(self.freq, self.Q)
        else:
            return 'pole {:f}\n'.format(self.freq)


class Zero:
    def __init__(self, freq, Q=None):
        self.freq = freq
        self.Q = Q

    def __repr__(self):
        if self.Q:
            return 'a c.c. zero pair at {:f} Hz with Q={:f}'.format(
                self.freq, self.Q)
        else:
            return 'a single zero at {:f} Hz'.format(self.freq)

    def compInstn(self):
        if self.Q:
            return 'zero {:f} {:e}\n'.format(self.freq, self.Q)
        else:
            return 'zero {:f}\n'.format(self.freq)
