import numpy as np
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
from matplotlib.ticker import FormatStrFormatter


def mag2db(arr, pow=False):
    if pow:
        return 10 * np.log10(arr)
    else:
        return 20 * np.log10(arr)


def plotTF(f, tf, mag_ax=None, phase_ax=None, dB=False, phase2freq=False,
           showUnstableRegion = False, **kwargs):
    '''
    Plots a transfer function.

    Inputs:
        f: the frequency at which the transfer function is evaluated [Hz]
        tf: the transfer function
        mag_ax: If not None, existing axis to plot the magnitude on
            (default: None)
        phase_ax: If not None, existing axis to plot the phase on
            (default: None)
        dB: If True, plots the magnitude in dB (default: False)
        phase2freq: If true, the transfer function tf/f is plotted instead
            of tf, which converts a phase TF to a frequency TF
            (default: False)
        **kwargs: any arguments (color, linestyle, label, etc.) to pass
            to the plot

    If mag_ax and phase_ax are None, then a new figure with the transfer
    function will be returned. However, if there are existing axes
    (for example from an existing TF plot) they can be passed to mag_ax
    and phase_ax and the new TF will be added to the existing axes. In this
    case no new figure is returned.

    Minimal labeling is done with this function. So, for example, to
    add a ylabel, title, and legend to a new TF
        fig = plotTF(ff, tf, label='legend label')
        mag_ax = fig.axes[0]
        mag_ax.set_xlabel('magnitude [W/m]')
        mag_ax.set_title('Plot Label')
        mag_ax.legend()
    Note that fig.axes[1] is phase_ax. A new TF can be plotted on this same
    figure by
        plotTF(ff, tf2, mag_ax=fig.axes[0], phase_ax=fig.axes[1], color='C1',
               label='second TF')
        mag_ax.legend()  # need to redraw legend
    '''
    if not(mag_ax and phase_ax):
        if (mag_ax is not None) or (phase_ax is not None):
            msg = 'If one of the phase or magnitude axes is given,'
            msg += ' the other must be given as well.'
            raise ValueError(msg)
        newFig = True
    else:
        newFig = False
    if newFig:
        fig = plt.figure()
        gs = gridspec.GridSpec(2, 1, height_ratios=[1, 1], hspace=0.05)
        mag_ax = fig.add_subplot(gs[0])
        phase_ax = fig.add_subplot(gs[1], sharex=mag_ax)
    if phase2freq:
        tf = tf/f
    if dB:
        mag_ax.semilogx(f, mag2db(np.abs(tf)), **kwargs)
        mag_ax.set_ylabel('magnitude [dB]')
        mag_ax.yaxis.set_major_formatter(FormatStrFormatter('%.0f'))
    else:
        mag_ax.loglog(f, np.abs(tf), **kwargs)
        mag_ax.set_ylabel('magnitude')
        magTF = np.abs(tf)
        # If the TF is close to being constant magnitude, increase ylims
        # in order to show y tick labels and avoid a misleading plot.
        if np.abs(np.max(magTF)/np.min(magTF)) < 10:
            # mag_ax.set_yscale('linear')
            mag_ax.set_ylim(np.mean(magTF)/10.1, np.mean(magTF)*10.1)

    mag_ax.set_xlim(min(f), max(f))
    phase_ax.set_ylim(-185, 185)
    # ticks = np.linspace(-180, 180, 7)
    ticks = np.arange(-180, 181, 45)
    phase_ax.yaxis.set_ticks(ticks)
    angTF = np.angle(tf, True)
    phase_ax.semilogx(f, angTF, **kwargs)
    phase_ax.set_ylabel('phase [deg]')
    phase_ax.set_xlabel('frequency [Hz]')
    plt.setp(mag_ax.get_xticklabels(), visible=False)
    mag_ax.grid('on', which='both')
    mag_ax.grid(alpha=0.2, which='minor')
    phase_ax.grid('on', which='both')
    phase_ax.grid(alpha=0.2, which='minor')

    if showUnstableRegion:
        btm = mag_ax.get_ylim()[0]
        if dB:
            unstab = mag2db([max(edge,btm) for edge in -2*np.cos(np.pi*angTF/180)])
        else:
            unstab = np.log10([max(edge,btm) for edge in -2*np.cos(np.pi*angTF/180)])
        mag_ax.fill_between(f,btm,unstab,where=unstab>=btm,facecolor='red',hatch='/', alpha=0.2)
    if newFig:
        return fig

def plotTFs(f, tfs, dB=False, showUnstableRegion = False, **kwargs):
    '''
    Plots multiple TFs in single plot.
    '''
    num_plots = len(tfs)
    colormap = plt.cm.Paired
    linestyles = ['-', '--',':', '-.']
    hatchstyles = ['/', '//', '+', 'x', 'o', '.']
    tfsMat = tfs.items()
    fig = plotTF(f, tfsMat[0][1], dB=dB, label = tfsMat[0][0],
                 ls=linestyles[0], **kwargs)
    for ii in range(1,num_plots):
        plotTF(f, tfsMat[ii][1], mag_ax=fig.axes[0], phase_ax=fig.axes[1],
               dB=dB, ls=linestyles[ii%4], label=tfsMat[ii][0], **kwargs)

    if showUnstableRegion:
        btm = fig.axes[0].get_ylim()[0]
        angTF = np.angle(tfsMat[0][1],True)
        if dB:
            unstab = mag2db([max(edge,btm) for edge in -2*np.cos(np.pi*angTF/180)])
        else:
            unstab = np.log10([max(edge,btm) for edge in -2*np.cos(np.pi*angTF/180)])
        fig.axes[0].fill_between(f,btm,unstab,where=unstab>=btm,
                                 facecolor='red', hatch=hatchstyles[0],
                                 label = tfsMat[0][0]+' Unst. Reg.',
                                 alpha=0.2)
        for ii in range(1,num_plots):
            angTF = np.angle(tfsMat[ii][1],True)
            if dB:
                unstab = mag2db([max(edge,btm) for edge in -2*np.cos(np.pi*angTF/180)])
            else:
                unstab = np.log10([max(edge,btm) for edge in -2*np.cos(np.pi*angTF/180)])
            fig.axes[0].fill_between(f,btm,unstab,where=unstab>=btm,
                                     facecolor='red', hatch=hatchstyles[ii%6],
                                     label = tfsMat[ii][0]+' Unst. Reg.',
                                     alpha=0.2)
    fig.axes[0].legend(ncol=2)  # need to redraw legend
    return fig

def plotMaxInput(f, maxInput, maxInpReason, **kwargs):
    '''
    Plots the maximum allowed input voltage at different frequencies and
    also gives the reason for restriction as different frequency regions.
    '''
    fig = plt.figure()
    fig.gca().semilogx(f,maxInput, **kwargs)
    fig.gca().set_ylabel('Maximum Input Voltage (V)')
    fig.gca().yaxis.set_major_formatter(FormatStrFormatter('%.0f'))
    fig.gca().set_xlim(min(f), max(f))
    fig.gca().grid('on', which='both')
    fig.gca().grid(alpha=0.2, which='minor')
    fig.gca().set_xlabel('frequency [Hz]')
    newylim = fig.gca().get_ylim()
    fig.gca().set_ylim([newylim[0]*1.15,newylim[1]*1.15])
    ArwYpos = (fig.gca().get_ylim()[1]*0.88 + fig.gca().get_ylim()[0]*0.12)
    TxtYpos = (fig.gca().get_ylim()[1]*0.85 + fig.gca().get_ylim()[0]*0.15)
    reasLabel = dict()
    startFreqs = np.sort(np.array(maxInpReason.values()))
    startFreqsInd = np.zeros(len(startFreqs)+1,int)
    for ii in range(len(startFreqs)):
        startFreqsInd[ii] = int(np.argmin(np.abs(f-startFreqs[ii])))
    startFreqsInd[-1]= int(len(f)-1)
    for reas,startfreq in maxInpReason.items():
        ind = np.argmin(np.abs(startfreq - startFreqs))
        startind = startFreqsInd[ind]
        stopind = startFreqsInd[ind+1]
        reasLabel[reas] = [f[startind],f[stopind],(f[startind]*0.9)+(f[stopind]*0.1)]
    for key,value in reasLabel.items():
        fig.gca().annotate('', xy=(value[0],ArwYpos), xycoords='data',
                           xytext=(value[1],ArwYpos), textcoords='data',
                           arrowprops={'arrowstyle': '|-|'})
        fig.gca().text(value[0],TxtYpos,key,fontsize='x-large')
    return fig

def plotASD(f, asd, dB=False, **kwargs):
    '''
    Plots an ASD

    Inputs:
        f: frequency at which the ASD is evaluated [Hz]
        asd: the ASD
        dB: If True, plots the magnitude in dB (default: False)
    **kwargs: any arguments (color, linestyle, label, etc.) to pass
        to the plot
    '''
    fig = plt.figure()
    if dB:
        fig.gca().semilogx(f, mag2db(asd), **kwargs)
        fig.gca().set_ylabel('ASD [dB]')
        fig.gca().yaxis.set_major_formatter(FormatStrFormatter('%.0f'))
    else:
        fig.gca().loglog(f, asd, **kwargs)
        fig.gca().set_ylabel('ASD [$V/Hz^{1/2}$]')
        # If the TF is close to being constant magnitude, increase ylims
        # in order to show y tick labels and avoid a misleading plot.
        if np.max(asd)/np.min(asd) < 10:
            fig.gca().set_ylim(np.mean(asd)/10.1, np.mean(asd)*10.1)
    fig.gca().set_xlim(min(f), max(f))
    fig.gca().grid('on', which='both')
    fig.gca().grid(alpha=0.2, which='minor')
    fig.gca().set_xlabel('frequency [Hz]')
    return fig

def plotASDs(f, asds, dB=False, **kwargs):
    '''
    Plots ASDs

    Inputs:
        f: frequency at which the ASD is evaluated [Hz]
        asds: a dictionary of asds, keys are LISO part names used as labels, values are noise ASDs in V/rtHz
        dB: If True, plots the magnitude in dB (default: False)
    **kwargs: any arguments (color, linestyle, label, etc.) to pass
        to the plot
    '''
    num_plots = len(asds)
    colormap = plt.cm.Paired
    linestyles = ['-', '--',':', '-.']

    fig = plt.figure()
    fig.gca().set_color_cycle([colormap(i) for i in np.linspace(0, 0.9, num_plots)])
    if dB:
        for ii, (key, asd) in enumerate(asds.iteritems()):
            fig.gca().semilogx(f, mag2db(asd), label=key, ls=linestyles[ii%4], **kwargs)
        fig.gca().set_ylabel('ASD [dB]')
        fig.gca().yaxis.set_major_formatter(FormatStrFormatter('%.0f'))
    else:
        for ii, (key, asd) in enumerate(asds.iteritems()):
            fig.gca().loglog(f, asd, label=key, ls=linestyles[ii%4], **kwargs)
        fig.gca().set_ylabel('ASD [$V/Hz^{1/2}$]')
        # If the TF is close to being constant magnitude, increase ylims
        # in order to show y tick labels and avoid a misleading plot.
        if np.max(asds.values())/np.min(asds.values()) < 10:
            fig.gca().set_ylim(np.mean(asd)/10.1, np.mean(asd)*10.1)
    fig.gca().set_xlim(min(f), max(f))
    fig.gca().grid('on', which='both')
    fig.gca().grid(alpha=0.2, which='minor')
    fig.gca().set_xlabel('frequency [Hz]')
    fig.gca().legend(ncol=3)
    return fig

def plotDomASDs(f, asds,CutoffRank=5, dB=False, **kwargs):
    '''
    Plots dominant ASDs

    Inputs:
        f: frequency at which the ASD is evaluated [Hz]
        asds: a dictionary of asds, keys are LISO part names used as labels, values are noise ASDs in V/rtHz
        CutoffRank: Any ASD which is never above the this rank will not be plotted
        dB: If True, plots the magnitude in dB (default: False)
    **kwargs: any arguments (color, linestyle, label, etc.) to pass
        to the plot
    '''
    valarray = np.array(asds.values())
    (ncurves,npoints) = np.shape(valarray)
    dominant = np.zeros((ncurves,npoints))
    for i in range(npoints):
        temp = np.argsort(valarray[:,i])
        ranks = np.empty_like(temp)
        ranks[temp] = np.arange(ncurves-1,-1,-1)
        dominant[:,i] = ranks
    Domasds = dict()
    for ii,(key,asd) in enumerate(asds.iteritems()):
        if min(dominant[ii,:])<CutoffRank:
            Domasds[key] = asd

    DomLabel = dict()
    same=0
    for frqind in range(npoints-1):
        if same is 0:
            startind=frqind
            for ii,(key,asd) in enumerate(asds.iteritems()):
                if dominant[ii,frqind]==0:
                    currLabel = key
                    break
            if dominant[ii,frqind+1]==0:
                same = 1
        else:
            if dominant[ii,frqind+1]==0:
                same = 1
            else:
                same = 0
                stopind = frqind
                DomLabel[currLabel] = [f[startind],f[stopind],(f[startind]**0.9)*(f[stopind]**0.1)]
    DomLabel[currLabel] = [f[startind],f[npoints-1],(f[startind]**0.9)*(f[npoints-1]**0.1)]

    num_plots = len(Domasds)
    colormap = plt.cm.Paired
    linestyles = ['-', '--',':', '-.']

    fig = plt.figure()
    fig.gca().set_color_cycle([colormap(i) for i in np.linspace(0, 0.9, num_plots)])
    if dB:
        for ii, (key, asd) in enumerate(Domasds.iteritems()):
            fig.gca().semilogx(f, mag2db(asd), label=key, ls=linestyles[ii%4], **kwargs)
        fig.gca().set_ylabel('ASD [dB]')
        fig.gca().yaxis.set_major_formatter(FormatStrFormatter('%.0f'))
        ArwYpos = (fig.gca().get_ylim()[1]**0.8)*(fig.gca().get_ylim()[0]**0.2)
        TxtYpos = (fig.gca().get_ylim()[1]**0.9)*(fig.gca().get_ylim()[0]**0.1)
    else:
        for ii, (key, asd) in enumerate(Domasds.iteritems()):
            fig.gca().loglog(f, asd, label=key, ls=linestyles[ii%4], **kwargs)
        fig.gca().set_ylabel('ASD [$V/Hz^{1/2}$]')
        # If the TF is close to being constant magnitude, increase ylims
        # in order to show y tick labels and avoid a misleading plot.
        if np.max(Domasds.values())/np.min(Domasds.values()) < 10:
            fig.gca().set_ylim(np.mean(asd)/10.1, np.mean(asd)*10.1)
        ArwYpos = (fig.gca().get_ylim()[1]*0.8 + fig.gca().get_ylim()[0]*0.2)
        TxtYpos = (fig.gca().get_ylim()[1]*0.9 + fig.gca().get_ylim()[0]*0.1)


    for key,value in DomLabel.items():
        fig.gca().annotate('', xy=(value[0],ArwYpos), xycoords='data',
                           xytext=(value[1],ArwYpos), textcoords='data',
                           arrowprops={'arrowstyle': '|-|'})
        fig.gca().text(value[2],TxtYpos,key,fontsize='xx-large')

    fig.gca().set_xlim(min(f), max(f))
    fig.gca().grid('on', which='both')
    fig.gca().grid(alpha=0.2, which='minor')
    fig.gca().set_xlabel('frequency [Hz]')
    fig.gca().legend(ncol=3)
    return fig
