'''
This is a python module for parsing and processing liso files using the
liso binary interpreter. It is intended to be used as a wrapper for liso for
those wanting to use python to handle simulations

Authors: Craig Cahillane and Andrew Wade (ccahilla@caltech.edu)
Date: circa Oct 2017
Modification log:

Modification author: Anchal Gupta (anchal@caltech.edu)
'''
import os
import numpy as np
import components
import pickle
import copy
from plotting import plotTF, plotASD, plotASDs, \
                     plotDomASDs, plotTFs, plotMaxInput


class Circuit:
    '''
    This runs a given liso .fil file and outputs to OutFile, reads it back
    in, and generates object with variables for frequency and complex
    transfer function

    opsys is the operating system, either linux or osx
    '''
    def __init__(self, opsys):
        # initializes the class variables
        if len(opsys) == 0:
            print('Assuming your system recognize liso command.')
            print('Choose linux or osx otherwise.')
            self.os = ''
        else:
            if opsys not in ['linux', 'osx']:
                msg = 'Unrecognized operating system {:s}.\n'.format(opsys)
                msg += 'Choose either linux or osx'
                raise ValueError(msg)
            self.os = '.'+opsys
        self.fname = None
        self.parts = {}
        self.ff = None
        self.TF = None
        self.TFinNode = None
        self.TFoutNode = None
        self.TFsrcZ = 50
        self.maxInput = None
        self.maxInpReason = {}
        self.noiseASDnode = None
        self.noiseASD = None
        self.allNoiseASDs = {}
        self.innoiseASD = None
        self.allInNoiseASDs = {}
        self.Zin = None
        self.ZinNode = None
        self.opStabs = {}
        self.indParam = []
        self.fitModelTF = None
        self.fittedTF = None
        self.fitEstimates = {}
        self.fitUncert = {}
        self.fitCorrMat = None
        self.redChiSqMod = None
        self.redChiSq = None
        self.objfname = None

    def getFil(self, fname):
        '''Displays the .fil file as it is written'''
        print '###############################'
        print 'Displaying', fname, ':'
        print '###############################'
        print
        # todo: put brackets around these print commands
        with open(self.InputFile, "r") as f:
            print f.read()

    def setInNode(self, inNode, srcZ=50):
        self.TFinNode = inNode
        self.TFsrcZ = srcZ

    def setOutNode(self, outNode):
        self.TFoutNode = outNode

    def parseFil(self, fname):
        '''Parses resistors, capacitors, inductors, opamps, voltage inputs,
        and voltage outputs from .fil file'''
        # To remember where the input parameters came from
        self.fname = fname.replace('.fil', '')
        with open(fname, "r") as f:
            # f.readlines() gives a list of lines in a file.  ignore comments
            for ii, line in enumerate(f.readlines()):
                # remove leading spaces, tabs, newlines, then trailing newlines
                line = line.lstrip().rstrip()
                if line == '':
                    continue
                if line[0] is '#':  # if line is a comment, continue
                    continue
                split = line.split(' ')  # divide up line based on spaces
                partType = split[0]
                if partType == 'r' or partType == 'c' or partType == 'l':
                    partName = split[1]
                    partValue = components.convert_SI_string_to_value(split[2])
                    partNodes = np.array([split[3], split[4]])
                    if partType == 'r':
                        self.addResistor(partName, partValue, *partNodes)
                    elif partType == 'c':
                        self.addCapacitor(partName, partValue, *partNodes)
                    elif partType == 'l':
                        self.addInductor(partName, partValue, *partNodes)
                elif partType == 'op':
                    partName = split[1]
                    partValue = split[2]
                    partNodes = np.array([split[3], split[4], split[5]])
                    self.addOpAmp(partName, partValue, *partNodes)
                # elif partType == 'uinput':
                #     partName = split[1]
                #     partNodes = split[1]
                #     partValue = 0
                # elif partType == 'uoutput':
                #     nextSplit = split[1].split(':')  # Split via colons now
                #     partName = nextSplit[0]
                #     partNodes = nextSplit[0]
                #     partValue = 0
                #     self.measType = nextSplit[1] + ':' + nextSplit[2]

                # self.parts[partName] = {'value': partValue,
                #                         'nodes': partNodes,
                #                         'type': partType}
        print '########################'
        print '#   Parsing complete   #'
        print '########################'
        print

    def displayParts(self):
        '''
        Prints a list of parts along with their values
        '''
        print('This circuit has the following parts:')
        for part in self.parts.values():
            print(part)

    def getPart(self, name):
        try:
            return self.parts[name]
        except KeyError:
            print('part {:s} does not exist'.format(name))

    def writeCircuitTopology(self, fname):
        '''
        Writes the circuit topology (all parts and their connections)
        to the file fname.
        Returns the handle to the file.
        '''
        f = open(fname, 'w')
        for part in self.parts.values():
            f.write(part.compInstn())
        return f

    def addResistor(self, name, value, node1, node2):
        self.parts[name] = components.Resistor(name, value, node1, node2)

    def addCapacitor(self, name, value, node1, node2):
        self.parts[name] = components.Capacitor(name, value, node1, node2)

    def addInductor(self, name, value, node1, node2):
        self.parts[name] = components.Inductor(name, value, node1, node2)

    def addOpAmp(self, name, typ, nodep, nodem, nodeout, **kwargs):
        self.parts[name] = components.OpAmp(name, typ, nodep, nodem,
                                            nodeout, **kwargs)

    def addMutualInductance(self, name, value, inductor1, inductor2):
        self.parts[name] = components.MutualInductance(name, value, inductor1,
                                                       inductor2)

    def addTransmissionLine(self, name, length, Z, C, R, G, fskin,
                            n3, n4, n5, n6):
        self.parts[name] = components.TransmissionLine(
            name, length, Z, Z, R, G, fskin, n3, n4, n5, n6)

    def computeTF(self, inNode, outNode, fmin,
                  fmax, npts, computeMaxInput=False, inNodeVolts=True):
        '''
        Computes a transfer function.
        '''
        # Remembering what inNode and outNode TF corresponds to.
        self.TFinNode = inNode
        self.TFoutNode = outNode
        f = self.writeCircuitTopology('buff.fil')
        if isinstance(inNode, basestring):
            if inNodeVolts:
                f.write('uinput {:s}\n'.format(inNode))
            else:  # handle case current input TI gain
                f.write('iinput {:s}\n'.format(inNode))
        else:
            if inNodeVolts:
                f.write('uinput {0} {1}\n'.format(inNode[0], inNode[1]))
            else:  # handle case current input TI gain
                f.write('iinput {0} {1}\n'.format(inNode[0], inNode[1]))
        CompInstn = ('uoutput {OUTNODE}:re:im\n'
                     'freq log {STFREQ} {SPFREQ} {NPOINT}\n').format(
                         OUTNODE=outNode,
                         STFREQ=fmin,
                         SPFREQ=fmax,
                         NPOINT=npts)
        f.write(CompInstn)
        if computeMaxInput:
            f.write('maxinput')
        f.close()

        # todo: do we also want to include 2>&1 to dump STDERR output?
        os.system('liso{:s} buff.fil buff.out /dev/null'.format(self.os))

        # Loads output file from liso, default skips lines starting with #.
        lisortn = np.loadtxt("buff.out")
        self.ff = lisortn[:, 0]  # get the frequency from the liso file
        # get the complex TF from liso file
        self.TF = lisortn[:, 1] + lisortn[:, 2]*1j
        # Calculating maximum allowed input
        if computeMaxInput:
            self.maxInput = lisortn[:, 3]
            f = open("buff.out", 'r')
            allLines = f.readlines()
            f.close()
            for ii, line in enumerate(allLines):
                if '#from' in line:
                    splits = line.split(' ')
                    self.maxInpReason[splits[8]] = float(splits[1])
                    if len(splits[2]) == 3:
                        self.maxInpReason[splits[8]] *= \
                         unit_prefix_to_multiplier(splits[2][0])
        # remove the buffering files
        os.remove('buff.fil')
        os.remove('buff.out')
        # os.remove('buff.gnu')

    def plotTF(self, mag_ax=None, phase_ax=None, dB=False,
               phase2freq=False, **kwargs):
        if self.TF is None:
            raise RuntimeError(
                'Need to compute a transfer function before plotting one!')
        fig = plotTF(self.ff, self.TF, mag_ax=mag_ax, phase_ax=phase_ax,
                     dB=dB, phase2freq=phase2freq, **kwargs)
        return fig

    def plotMaxInput(self, **kwargs):
        if self.maxInput is None:
            raise RuntimeError(
                'Need to run computeTF with computeMaxInput=True before'
                ' plotting the maximum input.'
            )
        fig = plotMaxInput(self.ff, self.maxInput, self.maxInpReason, **kwargs)
        return fig

    def computeNoise(self, inNode, evalNode,
                     fmin, fmax, npts, source_Z=50):
        '''
        This function will compute noise sum
        '''
        f = self.writeCircuitTopology('buff.fil')
        if isinstance(inNode, basestring):
            f.write('uinput {0:s} {1:s}\n'.format(inNode, str(source_Z)))
        else:
            f.write('uinput {0} {1} {2:s}\n'.format(inNode[0],
                                                    inNode[1],
                                                    str(source_Z)))
        f.write('noisy all\n')
        f.write('noise {:s} sum\n'.format(evalNode))
        f.write('freq log {:f} {:f} {:f}\n'.format(fmin, fmax, npts))
        f.close()
        #  os.system('liso.{OS} {INFIL} {OUTFIL} /dev/null'.format(
        #    OS=self.os, INFIL='buff.fil', OUTFIL='buff.out'))
        os.system('liso{:s} buff.fil buff.out /dev/null'.format(self.os))
        lisortn = np.loadtxt('buff.out')
        self.ff = lisortn[:, 0]
        self.noiseASD = lisortn[:, 1]
        os.remove('buff.fil')
        os.remove('buff.out')

    def computeAllNoises(self, inNode, evalNode,
                         fmin, fmax, npts, source_Z=50):
        '''
        This function will compute noise from all sources
        '''
        self.noiseASDnode = evalNode
        f = self.writeCircuitTopology('buff.fil')
        if isinstance(inNode, basestring):
            f.write('uinput {0:s} {1:s}\n'.format(inNode, str(source_Z)))
        else:
            f.write('uinput {0} {1} {2:s}\n'.format(inNode[0], inNode[1], str(source_Z)))
        f.write('noisy all\n')
        f.write('noise {:s} all\n'.format(evalNode))  # get all noises, not sum
        f.write('freq log {:f} {:f} {:f}\n'.format(fmin, fmax, npts))
        f.close()
        #  os.system('liso.{OS} {INFIL} {OUTFIL} /dev/null'.format(
        #    OS=self.os, INFIL='buff.fil', OUTFIL='buff.out'))
        os.system('liso{:s} buff.fil buff.out /dev/null'.format(self.os))
        lisortn = np.loadtxt('buff.out')
        (lines, noiseNum) = np.shape(lisortn)
        self.ff = lisortn[:, 0]

        # read in out file to get order of noises. resistors have one noise,
        # op amps have three.
        fname = 'buff.out'
        with open(fname, "r") as f:
            allLines = f.readlines()
            for ii, line in enumerate(allLines):
                if '#Noise is computed' in line:  # want next line after this
                    break
            line = allLines[ii+1]  # get next line with noises
            line = line.lstrip('# ').rstrip()
            splits = line.split(' ')
            for counter, split in enumerate(splits):
                # plus one since freq vec is 0th column
                self.allNoiseASDs[split] = lisortn[:, counter + 1]
        os.remove('buff.fil')
        os.remove('buff.out')

    def computeInNoise(self, inNode, evalNode, fmin, fmax, npts):
        '''
        This function will compute input reffered noise sum
        '''
        f = self.writeCircuitTopology('buff.fil')
        if isinstance(inNode, basestring):
            f.write('uinput {:s}\n'.format(inNode))
        else:
            f.write('uinput {0} {1}\n'.format(inNode[0], inNode[1]))
        f.write('noisy all\n')
        f.write('inputnoise {:s} sum\n'.format(evalNode))
        f.write('freq log {:f} {:f} {:f}\n'.format(fmin, fmax, npts))
        f.close()
        # os.system('liso.{OS} {INFIL} {OUTFIL} /dev/null'.format(
        #    OS=self.os, INFIL='buff.fil', OUTFIL='buff.out'))
        os.system('liso{:s} buff.fil buff.out /dev/null'.format(self.os))
        lisortn = np.loadtxt('buff.out')
        self.ff = lisortn[:, 0]
        self.innoiseASD = lisortn[:, 1]
        os.remove('buff.fil')
        os.remove('buff.out')

    def computeAllInNoises(self, inNode, evalNode, fmin, fmax, npts):
        '''
        This function will compute input reffered noise from all sources
        '''
        self.noiseASDnode = evalNode
        f = self.writeCircuitTopology('buff.fil')
        if isinstance(inNode, basestring):
            f.write('uinput {:s}\n'.format(inNode))
        else:
            f.write('uinput {0} {1}\n'.format(inNode[0], inNode[1]))
        f.write('noisy all\n')
        # get all noises, not sum
        f.write('inputnoise {:s} all\n'.format(evalNode))
        f.write('freq log {:f} {:f} {:f}\n'.format(fmin, fmax, npts))
        f.close()
        #  os.system('liso.{OS} {INFIL} {OUTFIL} /dev/null'.format(
        #    OS=self.os, INFIL='buff.fil', OUTFIL='buff.out'))
        os.system('liso{:s} buff.fil buff.out /dev/null'.format(self.os))
        lisortn = np.loadtxt('buff.out')
        (lines, noiseNum) = np.shape(lisortn)
        self.ff = lisortn[:, 0]

        # read in out file to get order of noises. resistors have one noise,
        # op amps have three.
        fname = 'buff.out'
        with open(fname, "r") as f:
            allLines = f.readlines()
            for ii, line in enumerate(allLines):
                if '#Noise is computed' in line:  # want next line after this
                    break
            line = allLines[ii+1]  # get next line with noises
            line = line.lstrip('# ').rstrip()
            splits = line.split(' ')
            for counter, split in enumerate(splits):
                # plus one since freq vec is 0th column
                self.allInNoiseASDs[split] = lisortn[:, counter + 1]
        os.remove('buff.fil')
        os.remove('buff.out')

    def plotNoise(self, dB=False, **kwargs):
        '''
        This function plots total noise sum ASD at output node selected while
        computing the noise using computeNoise()
        '''
        if self.noiseASD is None:
            raise RuntimeError(
                'Need to compute an ASD before plotting one!')
        fig = plotASD(self.ff, self.noiseASD, dB=dB, **kwargs)
        return fig

    def plotAllNoises(self, dB=False, **kwargs):
        '''
        This function plots contribution of all noise source ASDs at output node
        computed with computeAllNoises().
        If computeNoise() was also run (which computes sum of all noises), then
        the sum of all noises is also plotted.
        '''
        if not self.allNoiseASDs:
            raise RuntimeError(
                'Need to compute all ASDs before plotting them!'
                'Run computeAllNoises()')
        fig = plotASDs(self.ff, self.allNoiseASDs, dB=dB, **kwargs)
        if self.noiseASD is not None:
            fig.gca().loglog(self.ff,
                             self.noiseASD,
                             label='Sum {0}'.format(self.noiseASDnode),
                             color='k')
            fig.gca().legend(ncol=3)
        return fig

    def plotDomNoises(self, dB=False, CutoffRank=5, **kwargs):
        if not self.allNoiseASDs:
            raise RuntimeError(
                'Need to compute all ASDs before plotting them!'
                'Run computeAllNoises()')
        fig = plotDomASDs(self.ff, self.allNoiseASDs,
                          CutoffRank=CutoffRank, dB=dB, **kwargs)
        if self.noiseASD is not None:
            fig.gca().loglog(self.ff,
                             self.noiseASD,
                             label='Sum {0}'.format(self.noiseASDnode),
                             color='k')
            fig.gca().legend(ncol=3)
        return fig

    def plotInNoise(self, dB=False, **kwargs):
        if self.innoiseASD is None:
            raise RuntimeError(
                'Need to compute an ASD before plotting one!')
        fig = plotASD(self.ff, self.innoiseASD, dB=dB, **kwargs)
        return fig

    def plotAllInNoises(self, dB=False, **kwargs):
        if not self.allInNoiseASDs:
            raise RuntimeError(
                'Need to compute all ASDs before plotting them!'
                'Run computeAllNoises()')
        fig = plotASDs(self.ff, self.allInNoiseASDs, dB=dB, **kwargs)
        if self.innoiseASD is not None:
            fig.gca().loglog(self.ff,
                             self.innoiseASD,
                             label='Sum {0}'.format(self.noiseASDnode),
                             color='k')
            fig.gca().legend(ncol=3)
        return fig

    def plotDomInNoises(self, dB=False, CutoffRank=5, **kwargs):
        if not self.allInNoiseASDs:
            raise RuntimeError(
                'Need to compute all ASDs before plotting them!'
                'Run computeAllNoises()')
        fig = plotDomASDs(self.ff, self.allInNoiseASDs,
                          CutoffRank=5, dB=dB, **kwargs)
        if self.innoiseASD is not None:
            fig.gca().loglog(self.ff,
                             self.innoiseASD,
                             label='Sum {0}'.format(self.noiseASDnode),
                             color='k')
            fig.gca().legend(ncol=3)
        return fig

    def computeZin(self, inNode, fmin, fmax, npts):
        '''
        Computes a input impedance of the circuit
        '''
        self.ZinNode = inNode
        f = self.writeCircuitTopology('buff.fil')
        if isinstance(inNode, basestring):
            f.write('uinput {:s}\n'.format(inNode))
        else:
            f.write('uinput {0} {1}\n'.format(inNode[0], inNode[1]))
        CompInstn = ('zin :re:im\n'
                     'freq log {STFREQ} {SPFREQ} {NPOINT}\n').format(
                         STFREQ=fmin,
                         SPFREQ=fmax,
                         NPOINT=npts)
        f.write(CompInstn)
        f.close()

        # todo: do we also want to include 2>&1 to dump STDERR output?
        os.system('liso{:s} buff.fil buff.out /dev/null'.format(self.os))

        # Loads output file from liso, default skips lines starting with #.
        lisortn = np.loadtxt("buff.out")
        self.ff = lisortn[:, 0]  # get the frequency from the liso file
        # get the complex TF from liso file
        self.Zin = lisortn[:, 1] + lisortn[:, 2]*1j
        # remove the buffering files
        os.remove('buff.fil')
        os.remove('buff.out')
        # os.remove('buff.gnu')

    def plotZin(self, mag_ax=None, phase_ax=None, dB=False,
                phase2freq=False, **kwargs):
        if self.TF is None:
            raise RuntimeError(
                'Need to compute Zin before plotting one!')
        fig = plotTF(self.ff, self.Zin, mag_ax=mag_ax, phase_ax=phase_ax,
                     dB=dB, phase2freq=phase2freq, **kwargs)
        return fig

    def computeStability(self, inNode, fmin, fmax,
                         npts, source_Z=50.0, which_op='all'):
        '''
        This function will compute opamp stability. Stability of opamp is the
        open loop gain (OLG(f)) for the opamp feedback loopself. Input source
        is disabled with connecting source impedance to ground at the place.
        Opamp would be stable if |OLG(f) - 1|>1
        Roughly 45<OLG(f_ug)<315 (in degrees)
        Inputs:
            which_op: space separated list of opamps whose stability is to be
                      computed. Enter 'all' for selecting all opamps.
        '''
        f = self.writeCircuitTopology('buff.fil')
        if isinstance(inNode, basestring):
            f.write('uinput {:s} {:f}\n'.format(inNode, source_Z))
        else:
            f.write('uinput {0} {1}\n'.format(inNode[0], inNode[1], source_Z))
        # get all noises, not sum
        f.write('opstab {:s}:re:im\n'.format(which_op))
        f.write('freq log {:f} {:f} {:f}\n'.format(fmin, fmax, npts))
        f.close()
        #  os.system('liso.{OS} {INFIL} {OUTFIL} /dev/null'.format(
        #    OS=self.os, INFIL='buff.fil', OUTFIL='buff.out'))
        os.system('liso{:s} buff.fil buff.out /dev/null'.format(self.os))
        lisortn = np.loadtxt('buff.out')
        (lines, noiseNum) = np.shape(lisortn)
        self.ff = lisortn[:, 0]

        # read in out file to get order of noises. resistors have one noise,
        # op amps have three.
        fname = 'buff.out'
        with open(fname, "r") as f:
            allLines = f.readlines()
            for ii, line in enumerate(allLines):
                if '#OUTPUT' in line:  # want next line after this
                    break
            line = allLines[ii+1]  # get next line with noises
            line = line.replace(" Real Imag ", "")
            line = line.replace(" Real Imag", "")
            line = line.lstrip('# ').rstrip()
            splits = line.split(' ')
            for counter, split in enumerate(splits):
                self.opStabs[split] = \
                 lisortn[:, 2*counter + 1] + lisortn[:, 2*counter + 2]*1j
        os.remove('buff.fil')
        os.remove('buff.out')

    def plotOpStabs(self, dB=False, **kwargs):
        if not self.opStabs:
            raise RuntimeError(
                'Need to compute opamp stability!'
                'Run computeStability()')
        fig = plotTFs(self.ff, self.opStabs, showUnstableRegion=True,
                      dB=dB, **kwargs)
        return fig

    def optimizeCircuit(self, indParam, constraints, fitModel=None,
                        fitType='complex', minInput=None, minNoise=None,
                        update=False, weighting='rel', plotFit=True, **kwargs):
        '''
        This function optimizes the circuit instance to go near a model root
        instance or its own transfer function but remaining withing constraints
        and reducing noise and imporving input voltage window if asked for
        Inputs:
            indParam: Dictionary of independent parameters with key as name of
                      parameter and value as a list: [ lowerlimit, upperlimit]
                      constraints: Dependent parameters as equations of their
                      relations. List of strings where the dependent relation
                      is the only variable on lhs of equal. Only directly and
                      inversely propotional relations allowed.
            constraints: Dependent parameters as equations of their
                      relations. List of strings where the dependent relation
                      is the only variable on lhs of equal. Only directly and
                      inversely propotional relations allowed.
            fitModel: Root instance of an ideal transfer function. If not
                      provided, fit works with the present transfer function
                      and applies the constraints.
            fitType: String with following possible values:
                     'absolute': fit only magnitude of TF.
                     'complex': fit both magnitude and phase of TF.
            minInput: List for specifying constraint on the minimum expected
                      input signal. Format is [fromFreq, toFreq, value]
            minNoise: List for specifying constraint of minimum allowd noise.
                      Format is [node, fromFreq, toFreq, aim]
            update: If true, updates the current values of the Circuit instance
                    component values with the fitted values and write a new
                    input file with _fitted suffix for use later.
            weighting: Weighting of chi^2 function. String with options as
            liso:
                        'rel': w_i=1/|y_i|^2      'abs': w_i=1
                        'semi': w_i = 1/|y_i|
            plotFit: If true, plots the transfer functions of initial circuit,
                    model transfer function and fitted transfer function on
                    same graph.
        '''
        if fitModel is None:
            toFitFreq = self.ff
            toFitTF = self.TF
            if toFitTF is None:
                raise RuntimeError(
                    'Need to compute the transfer function of given circuit'
                    ' before optimizing for it when specific fitModel is'
                    ' not provided.')
        elif isinstance(fitModel, Root):
            if fitModel.TF is None:
                if self.ff is None:
                    raise RuntimeError(
                        'Need to compute transfer function of either fitModel'
                        ' or the circuit itself to get frequency data.'
                    )
                fitModel.computeTF(np.min(self.ff),
                                   np.max(self.ff), len(self.ff))
                print('Provided fit model does not have TF data.'
                      'Computing TF for fitModel using frequency data'
                      ' of circuit')
            toFitFreq = fitModel.ff
            toFitTF = fitModel.TF

        else:
            raise RuntimeError('Provided fitModel is not a valid Root object.')
        self.fitModelTF = toFitTF

        # Creating fit model data file.
        f = open('buff_TFdata.dat', 'w')
        if fitType is 'complex':
            for ii in range(len(toFitFreq)):
                f.write('{:f} {:f} {:f}\n'.format(toFitFreq[ii],
                        np.real(toFitTF[ii]), np.imag(toFitTF[ii])))
        elif fitType is 'absolute':
            for ii in range(len(toFitFreq)):
                f.write('{:f} {:f}\n'.format(toFitFreq[ii],
                        np.abs(toFitTF[ii])))
        else:
            raise RuntimeError(
                'fitType not recognized. Enter absolute or complex.'
            )
        f.close()

        self.indParam = list(indParam.keys())
        # Creating input file for fitting
        f = self.writeCircuitTopology('buff.fil')
        if self.TFinNode is None:
            raise RuntimeError(
                'Need to provide input node for circuit.'
                ' Either use setInNode() or run computeTF once.'
            )
        if self.TFoutNode is None:
            raise RuntimeError(
                'Need to provide output node for circuit.'
                ' Either use setOutNode() or run computeTF once.'
            )
        f.write('uinput {:s} {:f}\n'.format(self.TFinNode, self.TFsrcZ))
        f.write('uoutput {:s}\n'.format(self.TFoutNode))
        f.write('freq log {:f} {:f} {:f}\n'.format(min(toFitFreq),
                max(toFitFreq), len(toFitFreq)-1))

        # Apply constraints for optimizing the Dynamic Range
        applyDynConst(filehandle=f, minInput=minInput, minNoise=minNoise)
        # Write the fit instruction
        writeFitInst(f, indParam=indParam, constraints=constraints,
                     fitType=fitType, weighting=weighting)
        f.close()

        # todo: do we also want to include 2>&1 to dump STDERR output?
        os.system('liso{:s} buff.fil buff.out /dev/null'.format(self.os))
        # os.system('cp buff.fil buff_copy.fil') #Used during debugging
        os.remove('buff.fil')
        os.remove('buff_TFdata.dat')

        # Reading fitted estimates
        noip = len(indParam)            # No. of Independent Parameters
        nop = noip + len(constraints)   # Total no. of parameters
        with open('buff.out', "r") as f:
            allLines = f.readlines()
            for ii, line in enumerate(allLines):
                if 'Final' in line:
                    if line.split()[3] == 'inf':
                        raise RuntimeError(
                            'Fitting did not converge. Try to reduce the '
                            'degree of constraints.'
                        )
                if 'Correlation matrix' in line:
                    self.fitCorrMat = np.zeros((noip, noip))
                    matLines = allLines[ii+2:ii+2+noip]
                    for jj, row in enumerate(matLines):
                        splits = row.split()
                        for kk in range(jj-1):
                            self.fitCorrMat[jj, kk] = float(splits[kk+2])
                    for jj in range(noip):
                        for kk in range(noip-1, jj, -1):
                            self.fitCorrMat[jj, kk] = self.fitCorrMat[kk, jj]
                if 'Best parameter' in line:  # want next line after this
                    break
            for jj in range(ii+1, ii+nop+1, 1):
                estLine = allLines[jj].replace('#', '')
                splits = estLine.split()
                self.fitEstimates[splits[0]] = \
                    value_suffix_to_multiplier(splits[2])
                if len(splits) > 4:
                    self.fitUncert[splits[0]] = \
                        [splits[3], value_suffix_to_multiplier(splits[4])]
        # Reading chi square values
        chiSqModline = allLines[jj+2]
        chiSqModline = chiSqModline.replace('#', '')
        chiSqModline = chiSqModline.replace('Final chi^2=', '')
        self.redChiSqMod = float(chiSqModline.split()[0])/(len(toFitFreq)-noip)
        chiSqline = allLines[jj+3]
        if chiSqline.find('unmod') is not -1:
            chiSqline = chiSqline.replace('#', '')
            chiSqline = chiSqline.replace('unmod.=', '')
            self.redChiSq = float(chiSqline.split()[0])/(len(toFitFreq)-noip)
        # os.system('cp buff.out buff_copy.out') #Used during debugging
        os.remove('buff.out')

        print('Creating fitted Tranfer Function')
        self.createFitInputFile('buff_fit.fil')
        buffCircuit = Circuit(self.os.replace('.', ''))
        buffCircuit.parseFil('buff_fit.fil')
        buffCircuit.computeTF(self.TFinNode, self.TFoutNode, min(toFitFreq),
                              max(toFitFreq), len(toFitFreq)-1)
        self.fittedTF = buffCircuit.TF
        if self.ff is None:
            self.ff = buffCircuit.ff
        os.remove('buff_fit.fil')

        if update:
            self.update(self.fname+'_fitted.fil')

        if plotFit:
            return self.plotFit(**kwargs)

    def plotFit(self, **kwargs):
        '''
        plots the transfer functions of initial circuit,
        model transfer function and fitted transfer function on the
        same graph.
        '''
        if self.fittedTF is None:
            raise RuntimeError(
                'Need to do fitting before plotting fitted TF.'
                'Run optimizeCircuit()'
            )
        if self.TF is None:
            self.computeTF(self.TFinNode, self.TFoutNode, min(self.ff),
                           max(self.ff), len(self.ff)-1)
        fig = plotTFs(self.ff,
                      {'Initial TF':self.TF, 'Model TF':self.fitModelTF,'Fitted TF':self.fittedTF},
                      showUnstableRegion = False, **kwargs)
        chiSqText = 'Final red. $\chi^2$ = {:.3f}'.format(round(self.redChiSqMod,3))
        if self.redChiSq is not None:
            chiSqText = chiSqText+'\nUnmodified red. $\chi^2$ = {:.3f}'.format(round(self.redChiSq,3))
        mag_ax = fig.axes[0]
        mag_ax.text(0.4, 1, chiSqText,
                    horizontalalignment='left',
                    verticalalignment='top',
                    fontsize='xx-large', transform=mag_ax.transAxes)
        return fig

    def displayPartsRealValues(self, sequence='E12'):
        '''
        Prints a list of parts along with the nearest available real world value.
        '''
        print('This circuit has the following parts (Values chosen from {:s} sequence):'.format(sequence))
        partsCopy = copy.deepcopy(self.parts)
        for part in partsCopy.values():
            if not isinstance(part,components.OpAmp):
                part.value = (round_to_nearest_value(part.getValue(),sequence))
            print(part)

    def update(self,newFileName):
        '''
        This function updates the circuit instance with fit estimates and
        creates a new input file with component values from fit estimates.
        '''
        remParts = copy.deepcopy(self.fitEstimates)
        for partName, partValue in self.parts.iteritems():
            for param, fitValue in remParts.iteritems():
                if partName == param:
                    self.parts[partName].value = fitValue
                    del remParts[param]
                    break
        self.createFitInputFile(newFileName)

    def createFitInputFile(self, newFileName):
        '''
        This function creates an input file with the fitted component values.
        '''
        fr = open(self.fname+'.fil','r')
        allLines = fr.readlines()
        fr.close()
        if newFileName is not self.fname+'.fil':
            tempName = newFileName
        else:
            tempName = 'temp.fil'
        fw = open(tempName,'w')
        remParts = copy.deepcopy(self.fitEstimates)
        for line in allLines:
            writeLine = True
            for param, fitValue in remParts.iteritems():
                if line.find(param) is not -1:
                    oldValue = line.split()[2]
                    line=line.replace(oldValue,'{:f}'.format(fitValue))
                    fw.write(line)
                    writeLine=False
            if writeLine:
                fw.write(line)
        fw.close()
        if newFileName is self.fname+'.fil':
            os.remove(self.fname+'.fil')
            os.rename('temp.fil',newFileName)

class Root:
    '''
    This class will deal with LISO root mode.
    '''
    def __init__(self, opsys):
        if len(opsys) == 0:
            print('Assuming your system recognize liso command.\n')
            print('Choose linux or osx otherwise.')
            self.os=''
        else:
            if opsys not in ['linux', 'osx']:
                msg = 'Unrecognized operating system {:s}.\n'.format(opsys)
                msg += 'Choose either linux or osx'
                raise ValueError(msg)
                self.os = '.'+opsys
        self.poles = []
        self.zeroes = []
        self.factor = None
        self.ffactor = None
        self.delay = None
        self.TF = None
        self.objfname = None

    def addPole(self, freq, Q=None):
        self.poles.append(components.Pole(freq, Q))

    def addZero(self, freq, Q=None):
        self.zeroes.append(components.Zero(freq, Q))

    def setFactor(self, factor):
        self.factor = factor

    def setFfactor(self, ffactor):
        self.ffactor = ffactor

    def setDelay(self, delay):
        self.delay = delay

    def displayTF(self):
        '''
        Prints the TF definition
        '''
        print('This transfer function is defined with:')
        if len(self.poles):
            print('{:d} poles:'.format(len(self.poles)))
            for pole in self.poles:
                print(pole)
        if len(self.zeroes):
            print('{:d} zeroes:'.format(len(self.zeroes)))
            for zero in self.zeroes:
                print(zero)
        if self.factor:
            print('a gain of {:f}'.format(self.factor))
        if self.delay:
            print('a delay of {:f} s'.format(self.delay))
        if self.ffactor:
            print('a frequency scale factor of {:f}'.format(self.ffactor))

    def writeTF(self, fname):
        f = open(fname, 'w')
        for pole in self.poles:
            f.write(pole.compInstn())
        for zero in self.zeroes:
            f.write(zero.compInstn())
        if self.factor:
            f.write('factor {:f}\n'.format(self.factor))
        if self.ffactor:
            f.write('ffactor {:f}\n'.format(self.ffactor))
        if self.delay:
            f.write('delay {:f}\n'.format(self.delay))
        return f

    def computeTF(self, fmin, fmax, npts):
        f = self.writeTF('buff.fil')
        f.write('freq log {:f} {:f} {:f}\n'.format(fmin, fmax, npts))
        f.write('tfoutput re:im\n')
        f.close()

        #os.system('liso.{OS} {INFIL} {OUTFIL} /dev/null'.format(
        #    OS=self.os, INFIL='buff.fil', OUTFIL='buff.out'))
        os.system('liso{:s} buff.fil buff.out /dev/null'.format(self.os))
        lisortn = np.loadtxt('buff.out')
        self.ff = lisortn[:, 0]
        self.TF = lisortn[:, 1] + 1j*lisortn[:, 2]
        os.remove('buff.fil')
        os.remove('buff.out')

    def plotTF(self, mag_ax=None, phase_ax=None, dB=False,
               phase2freq=False, **kwargs):
        if self.TF is None:
            raise RuntimeError(
                'Need to compute a transfer function before plotting one!')
        fig = plotTF(self.ff, self.TF, mag_ax=mag_ax, phase_ax=phase_ax,
                     dB=dB, phase2freq=phase2freq, **kwargs)
        return fig

#***********************************************
# Overloading save and load for Circuit and Root
#***********************************************

def save(obj, objFileName = 'findName'):
    '''
    This function saves the object into a file.
    Please give '.cir' or '.rt' extension.
    If file name is not given, then function looks if the object knows the name
    given to its file. Otherwise, the objname in the instance is used to create
    a new file for it.
    '''
    pickle.HIGHEST_PROTOCOL = 3
    pickle.DEFAULT_PROTOCOL = 2
    if objFileName is 'findName':
        if isinstance(obj, Circuit):
            objFileName = obj.objfname
            if objFileName is None:
                objFileName = 'defaultCircuit.cir'
        elif isinstance(obj, Root):
            objFileName = obj.objfname
            if objFileName is None:
                objFileName = 'defaultRoot.rt'
        else:
            raise RuntimeError('Object is not Circuit or Root.'
                                   ' Do not know how to save this without filename.')
    if isinstance(obj, Circuit) or isinstance(obj, Root):
        obj.objfname = objFileName
    print('Saving object in file '+objFileName)
    with open(objFileName,'wb') as f:
        pickle.dump(obj,f,2)
        f.close()

def load(objFileName):
    '''
    This funciton creates a circuit or root object saved earlier with save function.
    '''
    pickle.HIGHEST_PROTOCOL = 3
    pickle.DEFAULT_PROTOCOL = 2
    with open(objFileName,'rb') as f:
        rtnobj = pickle.Unpickler(f).load()
        f.close()
        return rtnobj

#***********************************************
# Cluter reducing functions (Helper functions)
#***********************************************
def unit_prefix_to_multiplier(unit):
    '''
    This function reads the prefix letter in given word representing unit and
    gives the factor to multiply to come to SI unit. The function will assume
    there is a prefix present.
    Ex: 'mHz' -> 1.0e-3 'MHz' -> 1.0e6 etc
    '''
    return {
        'f' : 1.0e-15,
        'p' : 1.0e-12,
        'n' : 1.0e-9,
        'u' : 1.0e-6,
        'm' : 1.0e-3,
        'k' : 1.0e3,
        'M' : 1.0e6,
        'G' : 1.0e9,
        'T' : 1.0e12
    }[unit[0]]
def value_suffix_to_multiplier(value):
    '''
    This function reads the suffix letter in given word representing value and
    gives the factor to multiply to come to SI unit.
    Ex: '100m' -> 100*1.0e-3 '5n' -> 5n1.0e-9 etc
    '''
    try:
        return {
            'f' : 1.0e-15,
            'p' : 1.0e-12,
            'n' : 1.0e-9,
            'u' : 1.0e-6,
            'm' : 1.0e-3,
            'k' : 1.0e3,
            'M' : 1.0e6,
            'G' : 1.0e9,
            'T' : 1.0e12
        }[value[-1]]*float(value[0:-1])
    except:
        try:
            return float(value)
        except:
            print('Could not understand the output value '+value)

def applyDynConst(filehandle, minInput, minNoise):
    if isinstance(minInput,list):
        if len(minInput)==3:
            if not isinstance(minNoise,list):
                print('Warning: No minNoise read. Computed result might have '
                          'unrealistic component values.\n'
                          ' Please provide minNoise as a length 4 list'
                          ' in format [node, fromFreq, toFreq, aim] for better results.')
            elif not len(minNoise)==4:
                print('Warning: No minNoise read. Computed result might have '
                          'unrealistic component values.\n'
                          ' Please provide minNoise as a length 4 list'
                          ' in format [node, fromFreq, toFreq, aim] for better results.')
            filehandle.write('mininput {:f} {:f} {:f}\n'.format(minInput[0], minInput[1], minInput[2]))
    if isinstance(minNoise,list):
        if len(minNoise)==4:
            if not isinstance(minInput,list):
                print('Warning: No minInput read. Computed result might have '
                      'unrealistic operating window.\n'
                      ' Please provide minInput as a length 3 list'
                      ' in format [fromFreq, toFreq, value] for better results.')
            elif not len(minInput)==3:
                print('Warning: No minInput read. Computed result might have '
                      'unrealistic operating window.\n'
                      ' Please provide minInput as a length 3 list'
                      ' in format [fromFreq, toFreq, value] for better results.')
            filehandle.write('noisy all\n')
            filehandle.write('minnoise {:s} {:f} {:f} {:f}\n'.format(minNoise[0], minNoise[1], minNoise[2], minNoise[3]))

def writeFitInst(f, indParam, constraints, fitType, weighting):
    '''
    This function writes the fit instructions for liso
    It does not close the filehandle after the operation.
    Inputs:
        f: filehandle for the .fil file
        indParam: Dictionary of independent parameters with key as name of
                  parameter and value as a list: [ lowerlimit, upperlimit]
        constraints: Dependent parameters as equations of their relations. List
                     of strings where the dependent relation is the only
                     variable on lhs of equal. Only directly and inversely
                     propotional relations allowed.
        fitType: String with following possible values:
                 'absolute': fit only magnitude of TF.
                 'complex': fit both magnitude and phase of TF.
        weighting: Weighting of chi^2 function. String with options as liso:
                    'rel': w_i=1/|y_i|^2      'abs': w_i=1
                    'semi': w_i = 1/|y_i|
    '''
    if fitType is 'absolute':
        dataTypeStr = 'abs'
    elif fitType is 'complex':
        dataTypeStr = 'reim'
    else:
        raise RuntimeError(
            'fitType not recognized. Enter absolute or complex.'
        )
    f.write('fit buff_TFdata.dat {:s} {:s}\n'.format(dataTypeStr, weighting))
    remConstraints = copy.deepcopy(constraints)
    for param,limits in indParam.iteritems():
        f.write('param {:s} {:f} {:f}\n'.format(param, limits[0], limits[1]))
        for iterconstraint in remConstraints:
            cnstrnt = iterconstraint.replace(' ','')
            if cnstrnt.find(param) is not -1:
                Equalpos = cnstrnt.find('=')
                depParam = cnstrnt[0:Equalpos]
                if cnstrnt[Equalpos+1:].find('*') is not -1:
                    fac = unicode(cnstrnt[Equalpos+1:].replace(param,''),'utf-8')
                    fac = fac.replace('*','')
                    if fac.isnumeric():
                        f.write('pparam {:s} {:f}\n'.format(depParam, fac))
                    else:
                        f.write('pparam {:s}\n'.format(depParam, fac))
                elif cnstrnt[Equalpos+1:].find('/')is not -1:
                    Slashpos = cnstrnt.find('/')
                    fac = unicode(cnstrnt[Equalpos+1:].replace(param,''),'utf-8')
                    fac = fac.replace('/','')
                    if cnstrnt[Slashpos+1:].find(param) is not -1:
                        if fac.isnumeric():
                            f.write('rparam {:s} {:f}\n'.format(depParam, fac))
                        else:
                            f.write('rparam {:s}\n'.format(depParam, fac))
                    else:
                        if fac.isnumeric():
                            f.write('pparam {:s} {:f}\n'.format(depParam, 1/fac))
                        else:
                            f.write('pparam {:s}\n'.format(depParam, fac))
                else:
                    raise RuntimeError(
                        'Constraint not recognized. '
                        'Please use the any of the following format:\n'
                        'DepParam = IndepParam * factor\n'
                        'DepParam = factor * IndepParam\n'
                        'DepParam = IndepParam / factor\n'
                        'DepParam = factor / IndepParam\n'
                        'If factor is a number, it will be used.\n'
                        'If factor is anything else, it would be computed'
                        ' according to initial values')
                remConstraints.remove(iterconstraint)
def round_to_nearest_value(value, sequence = 'E12'):
    if sequence[0] is not 'E':
        raise RuntimeError(
                    'Unknown sequence type. Only E-series supported.'
        )
    seqNum = float(sequence[1:])
    if np.log2(seqNum/6)%1 != 0.0:
        raise RuntimeError(
                    'Unknown sequence type. Only E(6*2^n)-series supported.'
        )

    decNum = int(np.log10(value))-1
    normVal = value/(10**(decNum))
    if seqNum>24:
        sigDig=2
    else:
        sigDig=1

    closestPowNum = float(round(np.log10(normVal)*seqNum))
    return round(10**((closestPowNum/seqNum) + decNum),sigDig-decNum)
